import aiocache

from aiocache.cache import SimpleMemoryCache, RedisCache, MemcachedCache
from aiocache.utils import cached, multi_cached
from aiocache.serializers import DefaultSerializer
from aiocache.policies import DefaultPolicy


__all__ = (
    'cached',
    'multi_cached',
    'RedisCache',
    'SimpleMemoryCache',
    'MemcachedCache',
)

DEFAULT_CACHE = SimpleMemoryCache
DEFAULT_SERIALIZER = DefaultSerializer
DEFAULT_POLICY = DefaultPolicy
DEFAULT_NAMESPACE = ""
DEFAULT_KWARGS = {}


def set_defaults(cache, namespace=None, serializer=None, policy=None, **kwargs):
    aiocache.DEFAULT_CACHE = cache
    aiocache.DEFAULT_NAMESPACE = namespace or ""
    aiocache.DEFAULT_KWARGS = kwargs
    aiocache.DEFAULT_SERIALIZER = serializer or DefaultSerializer
    aiocache.DEFAULT_POLICY = policy or DefaultPolicy



import aiocache

from aiocache.backends import SimpleMemoryBackend, RedisBackend, MemcachedBackend


class BaseCache:
    """
    Base class that agregates the common logic for the different caches that may exist. Available
    options are:

    :param serializer: obj with :class:`aiocache.serializers.BaseSerializer` interface.
        Must implement ``loads`` and ``dumps`` methods.
    :param namespace: string to use as prefix for the key used in all operations of the backend.
    :param max_keys: int indicating the max number of keys to store in the backend. If not
        specified or 0, it's unlimited.
    """

    def __init__(self, serializer=None, namespace=None, *args, **kwargs):

        self._backend = self.get_backend(*args, **{**aiocache.DEFAULT_KWARGS, **kwargs})
        self._serializer = None
        self.serializer = serializer or self.get_default_serializer()
        self._policy = self.get_default_policy()
        self.namespace = namespace or aiocache.DEFAULT_NAMESPACE

    def get_backend(self, *args, **kwargs):
        raise NotImplementedError()

    def get_default_serializer(self):
        return aiocache.DEFAULT_SERIALIZER()

    def get_default_policy(self):
        return aiocache.DEFAULT_POLICY(self)

    @property
    def serializer(self):
        return self._serializer

    @serializer.setter
    def serializer(self, value):
        self._serializer = value
        new_encoding = getattr(self._serializer, "encoding", 'utf-8')
        self._backend.encoding = new_encoding

    @property
    def policy(self):
        return self._policy

    @policy.setter
    def policy(self, value):
        self._policy = value

    def set_policy(self, class_, *args, **kwargs):
        self._policy = class_(self, *args, **kwargs)

    async def add(self, key, value, ttl=None, dumps_fn=None):  # pragma: no cover
        """
        Stores the value in the given key with ttl if specified. Raises an error if the
        key already exists.
        :param key: str
        :param value: obj
        :param ttl: int the expiration time in seconds
        :param dumps_fn: callable alternative to use as dumps function
        :returns: True if key is inserted
        :raises: Value error if key already exists
        """
        dumps = dumps_fn or self._serializer.dumps
        ns_key = self._build_key(key)

        await self._policy.pre_set(key, value)
        await self._backend.add(ns_key, dumps(value), ttl)
        await self._policy.post_set(key, value)
        return True

    async def get(self, key, default=None, loads_fn=None):
        """
        Get a value from the cache. Returns default if not found.
        :param key: str
        :param default: obj to return when key is not found
        :param loads_fn: callable alternative to use as loads function
        :returns: obj loaded
        """

        loads = loads_fn or self._serializer.loads
        ns_key = self._build_key(key)

        await self._policy.pre_get(key)
        value = loads(await self._backend.get(ns_key))

        if value:
            await self._policy.post_get(key)

        return value or default

    async def multi_get(self, keys, loads_fn=None):
        """
        Get multiple values from the cache, values not found are Nones.

        :param keys: list of str
        :param loads_fn: callable alternative to use as loads function
        :returns: list of objs
        """
        loads = loads_fn or self._serializer.loads

        for key in keys:
            await self._policy.pre_get(key)

        values = await self._backend.multi_get([self._build_key(key) for key in keys])
        values = [loads(value) for value in values]

        for key in keys:
            await self._policy.post_get(key)

        return values

    async def set(self, key, value, ttl=None, dumps_fn=None):
        """
        Stores the value in the given key with ttl if specified
        :param key: str
        :param value: obj
        :param ttl: int the expiration time in seconds
        :param dumps_fn: callable alternative to use as dumps function
        :returns: True
        """
        dumps = dumps_fn or self._serializer.dumps
        ns_key = self._build_key(key)

        await self._policy.pre_set(key, value)
        await self._backend.set(ns_key, dumps(value), ttl)

        await self._policy.post_set(key, value)
        return True

    async def multi_set(self, pairs, ttl=None, dumps_fn=None):
        """
        Stores multiple values in the given keys.
        :param pairs: list of two element iterables. First is key and second is value
        :param ttl: int the expiration time of the keys in seconds
        :param dumps_fn: callable alternative to use as dumps function
        :returns: True
        """
        dumps = dumps_fn or self._serializer.dumps

        tmp_pairs = []
        for key, value in pairs:
            await self._policy.pre_set(key, value)
            tmp_pairs.append((self._build_key(key), dumps(value)))

        await self._backend.multi_set(tmp_pairs, ttl=ttl)

        for key, value in pairs:
            await self._policy.post_set(key, value)
        return True

    async def delete(self, key):
        """
        Deletes the given key.
        :param key: Key to be deleted
        :returns: int number of deleted keys
        """
        return await self._backend.delete(self._build_key(key))

    async def exists(self, key):
        """
        Check key exists in the cache.
        :param key: str key to check
        :returns: True if key exists otherwise False
        """
        return await self._backend.exists(self._build_key(key))

    async def raw(self, command, *args, **kwargs):
        """
        Send the raw command to the underlying client.

        :param command: str with the command.
        :returns: whatever the underlying client returns
        """
        return await self._backend.raw(command, *args, **kwargs)

    def _build_key(self, key):
        if self.namespace:
            return "{}{}".format(self.namespace, key)
        return key


class SimpleMemoryCache(BaseCache):
    """
    Cache implementation with the following components as defaults:
        - backend: :class:`aiocache.backends.SimpleMemoryBackend`
        - serializer: :class:`aiocache.serializers.DefaultSerializer`
        - policy: :class:`aiocache.policies.DefaultPolicy`
    """

    def get_backend(self, *args, **kwargs):
        return SimpleMemoryBackend(*args, **kwargs)


class RedisCache(BaseCache):
    """
    Cache implementation with the following components as defaults:
        - backend: :class:`aiocache.backends.RedisBackend`
        - serializer: :class:`aiocache.serializers.DefaultSerializer`
        - policy: :class:`aiocache.policies.DefaultPolicy`
    """

    def get_backend(self, *args, **kwargs):
        return RedisBackend(*args, **kwargs)

    def _build_key(self, key):
        if self.namespace:
            return "{}:{}".format(self.namespace, key)
        return key


class MemcachedCache(BaseCache):
    """
    Cache implementation with the following components as defaults:
        - backend: :class:`aiocache.backends.MemcachedBackend`
        - serializer: :class:`aiocache.serializers.DefaultSerializer`
        - policy: :class:`aiocache.policies.DefaultPolicy`
    """

    def get_backend(self, *args, **kwargs):
        return MemcachedBackend(*args, **kwargs)

    def _build_key(self, key):
        ns_key = super()._build_key(key)
        return str.encode(ns_key)



import inspect

import aiocache


def get_args_dict(fn, args, kwargs):
    defaults = {
        arg_name: arg.default for arg_name, arg in inspect.signature(fn).parameters.items()
        if arg.default is not inspect._empty
    }
    args_names = fn.__code__.co_varnames[:fn.__code__.co_argcount]
    return {**defaults, **dict(zip(args_names, args)), **kwargs}


def cached(ttl=0, key=None, key_attribute=None, cache=None, serializer=None, **kwargs):
    """
    Caches the functions return value into a key generated with module_name, function_name and args.

    In some cases you will need to send more args than just the ttl, cache and serializer.
    An example would be endpoint and port for the RedisCache. This extra args will be propagated
    to the cache class when instantiating.

    :param ttl: int seconds to store the function call. Default is 0
    :param key: str value to set as key for the function return. Takes precedence over
        key_attribute param.
    :param key_attribute: arg or kwarg name from the function to use as a key. If not passed,
        it will use module_name + function_name + args + kwargs
    :param cache: cache class to use when calling the ``set``/``get`` operations. Default is
        :class:``aiocache.SimpleMemoryCache``
    :param serializer: serializer instance to use when calling the ``serialize``/``deserialize``.
        Default is :class:``aiocache.serializers.DefaultSerializer``
    """
    cache = get_cache(cache=cache, serializer=serializer, **kwargs)

    def cached_decorator(fn):
        async def wrapper(*args, **kwargs):
            args_dict = get_args_dict(fn, args, kwargs)
            cache_key = key or args_dict.get(
                key_attribute, (fn.__module__ or 'stub') + fn.__name__ + str(args) + str(kwargs))
            if await cache.exists(cache_key):
                return await cache.get(cache_key)
            else:
                res = await fn(*args, **kwargs)
                await cache.set(cache_key, res, ttl=ttl)
                return res
        return wrapper
    return cached_decorator


def multi_cached(keys_attribute, key_builder=None, ttl=0, cache=None, serializer=None, **kwargs):
    """
    Only supports functions that return dict-like structures. This decorator caches each key/value
    of the dict-like object returned by the function.

    The decorated function must return a dict-like structure. Each key/value pair in the dict
    will be cached. If key_builder is passed, before storing the key will be transformed
    according to the output of the function

    If the attribute specified to be the key is empty, the cache will be ignored and the function
    will be called as expected.

    :param keys_attribute: arg or kwarg name from the function containing an iterable to use
        as keys.
    :param key_builder: Callable that allows to change the format of the keys before storing.
        Receives a dict with all the args of the function.
    :param ttl: int seconds to store the keys. Default is 0
    :param cache: cache class to use when calling the ``set``/``get`` operations. Default is
        :class:`aiocache.SimpleMemoryCache`
    :param serializer: serializer instance to use when calling the ``serialize``/``deserialize``.
        Default is :class:`aiocache.serializers.DefaultSerializer`
    """
    cache = get_cache(cache=cache, serializer=serializer, **kwargs)
    key_builder = key_builder or (lambda x, args_dict: x)

    def multi_cached_decorator(fn):
        async def wrapper(*args, **kwargs):
            partial_result = {}
            args_dict = get_args_dict(fn, args, kwargs)
            keys = args_dict[keys_attribute]
            cache_keys = [key_builder(key, args_dict) for key in keys]

            if len(keys) > 0:
                missing_keys = []
                values = await cache.multi_get(cache_keys)
                for key, value in zip(keys, values):
                    if value is not None:
                        partial_result[key] = value
                    else:
                        missing_keys.append(key)
                args_dict[keys_attribute] = missing_keys
            else:
                missing_keys = "all"

            if missing_keys:
                partial_result.update(await fn(**args_dict))
                if partial_result:
                    await cache.multi_set(
                        [(key_builder(
                            key, args_dict), value) for key, value in partial_result.items()],
                        ttl=ttl)

            return partial_result

        return wrapper
    return multi_cached_decorator


def get_cache(cache=None, serializer=None, policy=None, namespace=None, **kwargs):
    cache = cache or aiocache.DEFAULT_CACHE
    serializer = serializer or aiocache.DEFAULT_SERIALIZER()
    policy = policy or aiocache.DEFAULT_POLICY
    namespace = namespace or aiocache.DEFAULT_NAMESPACE

    c = cache(namespace=namespace, serializer=serializer, **{**aiocache.DEFAULT_KWARGS, **kwargs})
    c.set_policy(policy)
    return c
